/*
  Copyright 1995,1996 The University of Rhode Island and The Massachusetts
  Institute of Technology

  Portions of this software were developed by the Graduate School of
  Oceanography (GSO) at the University of Rhode Island (URI) in collaboration
  with The Massachusetts Institute of Technology (MIT).

  Access and use of this software shall impose the following obligations and
  understandings on the user. The user is granted the right, without any fee
  or cost, to use, copy, modify, alter, enhance and distribute this software,
  and any derivative works thereof, and its supporting documentation for any
  purpose whatsoever, provided that this entire notice appears in all copies
  of the software, derivative works and supporting documentation.  Further,
  the user agrees to credit URI/MIT in any publications that result from the
  use of this software or in any product that includes this software. The
  names URI, MIT and/or GSO, however, may not be used in any advertising or
  publicity to endorse or promote any products or commercial entity unless
  specific written permission is obtained from URI/MIT. The user also
  understands that URI/MIT is not obligated to provide the user with any
  support, consulting, training or assistance of any kind with regard to the
  use, operation and performance of this software nor to provide the user
  with any updates, revisions, new versions or "bug fixes".

  THIS SOFTWARE IS PROVIDED BY URI/MIT "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
  EVENT SHALL URI/MIT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
  DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE ACCESS, USE OR PERFORMANCE
  OF THIS SOFTWARE.
*/

// DODS-netCDF surrogate library functions for file handling. Note that with
// the preprocessor symbol LOCAL defined this library will access local files
// without using the DODS network services. Defining LOCAL will add support
// for local files through a slightly modifed version of the original netCDF
// library. This library has the same functionality as the original (it can
// access only files on a local hard disk) but has renamed entry points to
// avoid linker conflicts.
// 
// ReZa 10/13/94

// $Log: Dfile.cc,v $
// Revision 1.13  1998/11/11 18:14:28  jimg
// Fixed delete for of Connect objects for local files
//
// Revision 1.12  1998/03/20 00:31:00  jimg
// Changed quotes in error messages from double to single.
//
// Revision 1.11  1997/03/25 16:16:29  reza
// Fixed CVS duplicate log reports.
//
// Revision 1.10  1997/03/10 16:24:16  reza
// Added error object and upgraded to DODS core release 2.12.
//
// Revision 1.9  1996/09/19 16:09:38  jimg
// Removed bogus definition of the global rcode (which caused it to be defined
// twice when built without defining the preprocessor symbol LOCAL.
//
// Revision 1.8  1996/09/17 00:26:08  jimg
// Merged changes from a side branch which contained various changes from
// Reza and Charles.
// Removed ncdump and netexec since ncdump is now in its own directory and
// netexec is no longer used.
//
// Revision 1.7.2.2  1996/07/10 21:43:45  jimg
// Changes for version 2.06. These fixed lingering problems from the migration
// from version 1.x to version 2.x.
// Removed some (but not all) warning generated with gcc's -Wall option.
//
// Revision 1.7.2.1  1996/06/25 22:04:09  jimg
// Version 2.0 from Reza.
//
// Revision 1.7  1996/02/01 18:48:05  jimg
// Merged changes.
//
// Revision 1.6.2.3  1996/04/02 20:17:29  jimg
// GCC 2.6.3 does not handle template arguments which use the `*' operator; I
// changed the Connections<NCConnect *> definition to <NCConnectPtr>.
//
// Revision 1.6.2.2  1996/04/02 19:24:43  jimg
// Changed the `conns' variable so that now it is a global *pointer*, not an
// object. Each call that might add to the conns object must first check to see
// that it has been created. This was done because of link-time problems with
// static global objects.
// Fixed up some comment stuff.
// Added assert() calls for conns.
//
// Revision 1.6.2.1  1995/09/28  23:54:51  jimg
// Made the definitions of copts, cerr and cdf_routine_name extern when LOCAL
// is defined (these are defined in the file lnetcdf/lgolbdef.c).
//
// Revision 1.6  1995/07/09  21:33:33  jimg
// Added copyright notice.
//
// Revision 1.5  1995/06/29  15:42:04  jimg
// Fixed instances of delete which should have included brackets
//
// Revision 1.4  1994/12/22  04:51:39  reza
// Updated to use DODS new named dimension capability.
//
// Revision 1.3  1994/12/08  19:01:34  reza
// Added support for local netCDF file access through the original netCDF code.
//
// Revision 1.2  1994/11/23  21:05:58  reza
// First working version.
//
// Revision 1.1  1994/11/03  04:37:58  reza
// Preliminary version of netCDF -> DODS library.

#include "config_dap.h"

static char rcsid[] __unused__ ={"$Id: Dfile.cc,v 1.13 1998/11/11 18:14:28 jimg Exp $"};

#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>
#include <String.h>

#include "Dnetcdf.h"

#include "Connect.h"
#include "NCConnect.h"

// Make these symbols external so as to avoid link-time conflicts with their
// definitions in lfile.c (the localized version of the original
// implementation of netcdf.

#ifdef LOCAL
extern char *cdf_routine_name;
extern int ncopts;
extern int ncerr;
#else
char *cdf_routine_name;
int ncopts = NC_VERBOSE;
int ncerr;
#endif

int rcode;                  

// Remove the static global `conns' and replace it with a global pointer that
// client functions will test and initialize before using. This change will
// (hopefully) allow users to link programs with compilers other than g++.
// jhrg 4/2/96

#ifdef NEVER
Connections<NCConnectPtr>  conns(MAX_NC_OPEN);
#endif
Connections<NCConnectPtr> *conns = 0;

int
nccreate(const char *path, int cmode)
{
    int id;
    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if(c->is_local()) {  
	rcode = lnccreate(path, cmode);
	if (rcode != -1) {
	    c->ncid() = rcode;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    id = conns->add_connect(c);
	    return id;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    cdf_routine_name = "nccreate" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    //tmp_out    delete c;
    return(-1);
}


int
ncopen(const char *urlpath, int cmode)
{
    int id;
    String path = urlpath;
    bool status = true;

    String ce = path.after("?");
    if (path.contains("?"))
	path = path.before("?");

    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if (c->is_local()) {  
	rcode = lncopen((const char *)path, cmode);
	if (rcode != -1) {
	    c->ncid() = rcode;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    id = conns->add_connect(c);
	    return id;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    cdf_routine_name = "ncopen";

    if (cmode != NC_NOWRITE) {    
	NCadvise(NC_EPERM, 
		 "Bad file mode, Only read access allowed over the network");
	status = false;
	goto exit;
    }

    if (!c->request_das()) {
	NCadvise(NC_EINVAL, "Error in getting DAS from `%s'",
		 (const char *)path);
	status = false;
	goto exit;
    }

    if (!c->request_dds()) {
	NCadvise(NC_EINVAL, "Error in getting DDS from `%s'",
		 (const char *)path);
	status = false;
	goto exit;
    }

    if (!c->dds().parse_constraint(ce)) {
	NCadvise(NC_EINVAL, "Error in parsing constarint from `%s'",
		 (const char *)path);
	status = false;
	goto exit;
    }

    if (!c->request_Cdds(ce)) {
	NCadvise(NC_EINVAL, "Error in getting CDDS from `%s'",
		 (const char *)path);
	status = false;
	goto exit;
    }

    // well now we can extract dim. and attr. info. into tables
    c->parse_das_loc();
    c->parse_dims();

#ifdef TEST
    for (int i=0 ; i<7 ; i++)
	cerr << i << "-" << c->das_loc(i) << endl;
    c->das().print();
    c->dds().print();
#endif

exit:
    if (status == true) {
	if (!conns)
	    conns = new Connections<NCConnect *>(MAX_NC_OPEN);

	return(conns->add_connect(c)) ;
    }
    else {
      //tmp_out	delete c;
	return -1;
    }
}


int
ncclose(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;
#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncclose((*conns)[cdfid]->ncid());

	delete (*conns)[cdfid];
	conns->del_connect(cdfid);

	return rcode;
    }
#endif

    cdf_routine_name = "ncclose" ;

#ifdef TEST
    (*conns)[cdfid]->das().print();
    (*conns)[cdfid]->dds().print();
#endif

    delete (*conns)[cdfid];

    conns->del_connect(cdfid);
    
    return(0) ;
}


int
ncsync(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncsync((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncsync" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    return(-1);
}


int
ncabort(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncabort((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif
    
    cdf_routine_name = "ncabort" ;

    rcode = ncclose(cdfid);

    return rcode;
}


/* 
 * Deprecated function ;
 */
int
ncnobuf(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

    cdf_routine_name = "ncnobuf" ;

    /* NOOP */

    return(0);
}


int
ncredef(int cdfid)
{
    assert(conns);
    
    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {  
	rcode = lncredef((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncredef" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;
    return(-1);
}


int
ncendef( int cdfid )
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) { 
	rcode = lncendef((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncendef" ;

    return(0);
}


int
ncsetfill(int cdfid, int fillmode)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {  
	rcode = lncsetfill((*conns)[cdfid]->ncid(), fillmode);
	return rcode;
    }
#endif

    cdf_routine_name = "ncsetfill" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    return(-1);
}



